/*-
 * $Id: rr-message-start.c,v 1.23 2003/02/05 09:03:26 jonas Exp $
 *
 * See the file LICENSE for redistribution information. 
 * If you have not received a copy of the license, please contact CodeFactory
 * by email at info@codefactory.se, or on the web at http://www.codefactory.se/
 * You may also write to: CodeFactory AB, SE-903 47, Ume, Sweden.
 *
 * Copyright (c) 2002, 2003 Jonas Borgstrm <jonas@codefactory.se>
 * Copyright (c) 2002, 2003 Daniel Lundin   <daniel@codefactory.se>
 * Copyright (c) 2002, 2003 CodeFactory AB.  All rights reserved.
 */

#include <librr/rr.h>
#include <string.h>

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include <stdio.h>

struct _RRMessageStartPriv {
	gint number;
	GSList *in_list;
	GSList *out_list;

	RRChannel *done_channel;
	RRWaiter *waiter;
	GError *client_init_error;
	
	gchar *server_name;
};

static GObjectClass *parent_class = NULL;
static void finalize (GObject *object);
static gboolean process_frame (RRMessage *message, RRFrame *frame, 
				  GError **error);

static RRFrame *get_frame (RRMessage *stat, gsize max_size);

static void
free_item (gpointer data, gpointer user_data)
{
	RRStartItem *item = (RRStartItem *)data;
	g_free (item->piggyback);
	g_free (item);
}

static void
finalize (GObject *object)
{
	RRMessageStart *start = (RRMessageStart *)object;
	RRMessageStartPriv *priv = start->priv;
	
	rr_waiter_free (priv->waiter);

	g_slist_foreach (priv->in_list, free_item, 0);
	g_slist_foreach (priv->out_list, (GFunc)g_object_unref, 0);
	g_slist_free (priv->in_list);
	g_slist_free (priv->out_list);

	g_free (priv->server_name);
	if (priv->client_init_error)
		g_error_free (priv->client_init_error);

	g_free (priv);
	parent_class->finalize (object);
}

static void
rr_message_start_init (GObject *object)
{
	RRMessageStart *start = (RRMessageStart *)object;
	RRMessageStartPriv *priv;
	RRMessage *msg = (RRMessage *)object;

	priv = g_new0 (RRMessageStartPriv, 1);
	start->priv = priv;

	priv->waiter     = rr_waiter_new ();
	msg->type = RR_FRAME_TYPE_MSG;
}

static void
rr_message_start_class_init (GObjectClass *klass)
{
	RRMessageClass *msg_class = (RRMessageClass *)klass;

	klass->finalize = finalize;
	msg_class->get_frame = get_frame;
	msg_class->process_frame = process_frame;

	parent_class = g_type_class_peek_parent (klass);
}

GType 
rr_message_start_get_type (void)
{
	static GType rr_type = 0;

	if (!rr_type) {
		static GTypeInfo type_info = {
			sizeof (RRMessageStartClass),
			NULL,
			NULL,
			(GClassInitFunc) rr_message_start_class_init,
			NULL,
			NULL,
			sizeof (RRMessageStart),
			16,
			(GInstanceInitFunc) rr_message_start_init
		};
		rr_type = g_type_register_static (RR_TYPE_MESSAGE, 
						  "RRMessageStart", 
						  &type_info, 0);
	}
	return rr_type;
}


RRMessageStart *
rr_message_start_new (gint number, const gchar *server_name)
{
	RRMessageStart *start;
	RRMessageStartPriv *priv;

	start = g_object_new (RR_TYPE_MESSAGE_START, NULL);
	priv = start->priv;

	priv->number = number;
	if (server_name)
		priv->server_name = g_strdup (server_name);

	return start;
}

GSList *
rr_message_start_get_channel_list (RRMessageStart *start)
{
	g_return_val_if_fail (RR_IS_MESSAGE_START (start), NULL);

	return start->priv->in_list;
}

void
rr_message_start_add_channel (RRMessageStart *start,
			      RRConnection *conn,
			      GType type,
			      gpointer config_data)
{
	RRProfileRegistry *profreg;
	RRMessageStartPriv *priv = start->priv;
	RRChannel *channel;
	GError *error = NULL;
	gpointer global_config = NULL;

	channel = g_object_new (type, NULL);
	channel->id = priv->number;
	channel->instance_config = config_data;

	profreg = rr_connection_get_profile_registry (conn);
	if (profreg) {
		global_config = rr_profile_registry_get_global_config (profreg,
								       type);
	}

	channel->global_config = global_config;

	rr_channel_set_connection (channel, conn);
	/* Give the channel a chance to generate a piggyback */
	if (!rr_channel_client_init (channel, &error)) {
		
		if (error) {
			if (priv->client_init_error)
				g_error_free (priv->client_init_error);
			priv->client_init_error = error;
		}
		rr_debug1 ("message_start::add_channel "
			   "rr_channel_client_init failed: %s, %s", 
			   G_OBJECT_TYPE_NAME (G_OBJECT (channel)),
			   error->message);		
		g_object_unref (G_OBJECT (channel));
	}
	else
		priv->out_list = g_slist_append (priv->out_list, channel);
}

static void
append_channel (gpointer data, gpointer user_data)
{
	RRChannel *channel = data;
	GString *str = (GString *)user_data;
	const gchar *uri, *piggyback;

	g_return_if_fail (RR_IS_CHANNEL (channel));
	
	uri = rr_channel_get_uri (G_OBJECT_TYPE (G_OBJECT (channel)));
	piggyback = rr_channel_get_piggyback (channel);

	if (piggyback) {
		g_string_append_printf (str, 
					"   <profile uri='%s'>\r\n"
					"      <![CDATA[%s]]>\r\n"
					"   </profile>\r\n",
					uri, piggyback);
	}
	else
		g_string_append_printf (str, "   <profile uri='%s' />\r\n", uri);
}

static gboolean
process_frame (RRMessage *message, RRFrame *frame, GError **error)
{
	RRMessageStart *start;
	RRMessageStartPriv *priv;
	RRConnection *conn;
	xmlDocPtr doc;
	xmlNodePtr node;
	gboolean ret;
	gchar *server_name;
	xmlChar *str;
	const gchar *data;
	gint len;

	g_return_val_if_fail (RR_IS_MESSAGE_START (message), FALSE);
	g_return_val_if_fail (RR_IS_FRAME (frame), FALSE);
	g_return_val_if_fail (RR_IS_CHANNEL (message->channel), FALSE);
	g_return_val_if_fail (RR_IS_CONNECTION (message->channel->connection), 
			      FALSE);

	conn = message->channel->connection;
	start = RR_MESSAGE_START (message);
	priv = start->priv;

	data = rr_frame_mime_get_body (frame);
	len  = rr_frame_mime_get_body_size (frame);

	doc = xmlParseMemory (data, len);
	if (doc == NULL) {
		g_set_error (error, RR_ERROR, RR_BEEP_CODE_SYNTAX_ERROR, "Invalid start message.");
		return FALSE;
	}

	node = xmlDocGetRootElement (doc);

	if (strcmp (node->name, "start") != 0)
		goto err;

	if ((str = xmlGetProp (node, "number")) == NULL)
		goto err;
	
	priv->number = atoi (str);
	xmlFree (str);

	if ((server_name = xmlGetProp (node, "serverName"))) {

		priv->server_name = g_strdup (server_name);
		xmlFree (server_name);
	}
	
	node = node->children;

	while (node) {
		xmlChar *uri, *piggyback = NULL;
		GType type;
		RRStartItem *item;
		RRProfileRegistry *profreg;

		/* Ignore none element nodes */
		while (node && node->type != XML_ELEMENT_NODE)
			node = node->next;
		if (node == NULL)
			break;

		if (strcmp (node->name, "profile") != 0)
			goto err;

		if ((uri = xmlGetProp (node, "uri")) == NULL)
			goto err;

		profreg = rr_connection_get_profile_registry (conn);
		type = rr_profile_registry_lookup_by_uri (profreg, uri);
		xmlFree (uri);
		if (type == G_TYPE_INVALID) {
			node = node->next;
			continue;
		}
		/* Extract the piggyback data if it exists */
		if (node->children) {
			xmlNode *child, *cdata_node = NULL;

			child = node->children;
			while (child) {
				if (child->type == XML_CDATA_SECTION_NODE &&
				    child->content) {
					cdata_node = child;
					break;
				}
				child = child->next;
			}
			if (cdata_node && child->content)
				piggyback = cdata_node->content;
			else if (node->children->type == XML_TEXT_NODE &&
				 node->children->content)
				piggyback = node->children->content;
		}
		item = g_new (RRStartItem, 1);
		item->type = type;
		item->piggyback = g_strdup (piggyback);
		priv->in_list = g_slist_append (priv->in_list, item);
		node = node->next;
	}
	ret = TRUE;
	goto end;
 err:
	g_set_error (error, RR_ERROR, RR_BEEP_CODE_PARAM_ERROR, 
		     RR_GERROR_DEFAULT_MESSAGE); 
	ret = FALSE;
 end:
	xmlFreeDoc (doc);

	return ret;
}

static RRFrame *
get_frame (RRMessage *message, gsize max_size)
{
	RRMessageStart *start = RR_MESSAGE_START(message);
	RRMessageStartPriv *priv = start->priv;
	RRFrame *frame;
	GString *str = g_string_new (RR_BEEP_MIME_HEADER);

	g_assert (RR_IS_MESSAGE (start));

	if (priv->server_name)
		g_string_append_printf (str, "<start number='%d' serverName='%s'>\r\n", 
					priv->number, priv->server_name);
	else
		g_string_append_printf (str, "<start number='%d'>\r\n", 
					priv->number);

	g_slist_foreach (priv->out_list, append_channel, str);

	g_string_append (str, "</start>\r\n");

	if (str->len > max_size) {

		g_string_free (str, TRUE);
		return NULL;
	}

	frame = rr_frame_new (RR_FRAME_TYPE_MSG, 
			      message->channel->id, FALSE, message->msgno,
			      str->len, 0, str->str, TRUE);

	g_string_free (str, FALSE);

	return frame;
}

void
rr_message_start_done (RRMessageStart *start, RRChannel *channel,
		       GError *error)
{
	g_return_if_fail (RR_IS_MESSAGE_START (start));

	start->priv->done_channel = channel;
	rr_waiter_set_done (start->priv->waiter, error);
}

RRChannel *
rr_message_start_wait_for_reply (RRMessageStart *start, GError **error)
{
	RRMessageStartPriv *priv;
	
	g_return_val_if_fail (RR_IS_MESSAGE_START (start), NULL);

	priv = start->priv;

	rr_waiter_wait (priv->waiter, error);
	return priv->done_channel;
}

/**
 * rr_message_start_get_server_name:
 * @start: A #RRMessageStart.
 * 
 * Returns the "serverName" attribute for the first successful "start" element
 * received by a BEEP peer is meaningful for the duration of the BEEP
 * session.
 * 
 * Return value: the %serverName.
 **/
const gchar *
rr_message_start_get_server_name (RRMessageStart *start)
{
	g_return_val_if_fail (RR_IS_MESSAGE_START (start), NULL);

	return start->priv->server_name;
}

gboolean
rr_message_start_empty_request_p (RRMessageStart *start,
				  GError **error)
{
	RRMessageStartPriv *priv;
	g_return_val_if_fail (RR_IS_MESSAGE_START (start), TRUE);

	priv = start->priv;

	if (priv->out_list == NULL) {
		g_propagate_error (error, 
				   priv->client_init_error);
		priv->client_init_error = NULL;
		return TRUE;
	}
	return FALSE;
}

gint
rr_message_start_get_number (RRMessageStart *start)
{
	g_return_val_if_fail (RR_IS_MESSAGE_START (start), -1);

	return start->priv->number;
}

static gint 
find_channel_helper (gconstpointer data, gconstpointer user_data)
{
	RRChannel *channel = RR_CHANNEL (data);
	GType type = GPOINTER_TO_UINT (user_data);
	
	return G_OBJECT_TYPE (channel) != type;
}

RRChannel *
rr_message_start_find_channel (RRMessageStart *start,
			       GType type)
{
	GSList *list;

	g_return_val_if_fail (RR_IS_MESSAGE_START (start), NULL);

	list = g_slist_find_custom (start->priv->out_list,
				    GUINT_TO_POINTER (type),
				    find_channel_helper);

	return list ? list->data : NULL;
}
