/*-
 * $Id: rr-waiter.c,v 1.4 2003/05/27 10:13:14 jonas Exp $
 *
 * See the file LICENSE for redistribution information. 
 * If you have not received a copy of the license, please contact CodeFactory
 * by email at info@codefactory.se, or on the web at http://www.codefactory.se/
 * You may also write to: CodeFactory AB, SE-903 47, Ume, Sweden.
 *
 * Copyright (c) 2002, 2003 Jonas Borgstrm <jonas@codefactory.se>
 * Copyright (c) 2002, 2003 Daniel Lundin   <daniel@codefactory.se>
 * Copyright (c) 2002, 2003 CodeFactory AB.  All rights reserved.
 */

#include "rr-waiter.h"

struct _RRWaiter {
	gboolean done;
	GError  *error;
	GMutex  *mutex;
	GCond   *cond;
	gpointer user_data;
};

/**
 * rr_waiter_new:
 * 
 * Creates a new #RRWaiter instance.
 * 
 * Return value: a new #RRWaiter instance.
 **/
RRWaiter *
rr_waiter_new ()
{
	RRWaiter *waiter;

	waiter = g_new (RRWaiter, 1);
	waiter->done      = FALSE;
	waiter->error     = NULL;
	waiter->mutex     = g_mutex_new ();
	waiter->cond      = g_cond_new ();

	return waiter;
}

/**
 * rr_waiter_free:
 * @waiter: a #RRWaiter instance
 * 
 * free the resources allocated by @waiter.
 **/
void
rr_waiter_free (RRWaiter *waiter)
{
	g_return_if_fail (waiter != NULL);
	
	if (waiter->error) {
		g_error_free (waiter->error);
	}
	g_mutex_free (waiter->mutex);
	g_cond_free  (waiter->cond);
	g_free (waiter);
}

/**
 * rr_waiter_set_done:
 * @waiter: a #RRWaiter instance
 * @error: an error message or %NULL
 * 
 * Ublock all waiting threads.
 **/
void
rr_waiter_set_done (RRWaiter *waiter, GError *error)
{
	g_return_if_fail (waiter != NULL);

	g_mutex_lock (waiter->mutex);
	waiter->done = TRUE;
	if (error) {
		if (waiter->error) {
			g_error_free (waiter->error);
		}
		waiter->error = g_error_copy (error);
	}
	g_cond_broadcast (waiter->cond);
	g_mutex_unlock   (waiter->mutex);
}

/**
 * rr_waiter_wait:
 * @waiter: a #RRWaiter instance
 * @error: a location to store eventual error messages.
 * 
 * block until some other thread unlocks with a rr_waiter_set_done call.
 * 
 * Return value: TRUE if the "error" argument to rr_waiter_set_done is %NULL.
 **/
gboolean
rr_waiter_wait (RRWaiter *waiter, GError **error)
{
	g_return_val_if_fail (waiter != NULL, FALSE);

	g_mutex_lock (waiter->mutex);
	while (waiter->done == FALSE) {
		g_cond_wait (waiter->cond, waiter->mutex);
	}
	if (waiter->error) {
		g_propagate_error (error, waiter->error);
		waiter->error = NULL;
		g_mutex_unlock (waiter->mutex);
		return FALSE;
	}
	g_mutex_unlock (waiter->mutex);
	return TRUE;
}

/**
 * rr_waiter_set_user_data:
 * @waiter: a #RRWaiter instance
 * @user_data: custom user data
 * 
 * store @user_data within the #RRWaiter instance.
 **/
void
rr_waiter_set_user_data (RRWaiter *waiter, gpointer user_data)
{
	g_return_if_fail (waiter != NULL);

	g_mutex_lock   (waiter->mutex);
	waiter->user_data = user_data;
	g_mutex_unlock (waiter->mutex);
}

/**
 * rr_waiter_get_user_data:
 * @waiter: a #RRWaiter instance
 * 
 * Retrieve the user data stored within the #RRWaiter instance
 * 
 * Return value: user data
 **/
gpointer
rr_waiter_get_user_data (RRWaiter *waiter)
{
	gpointer data;

	g_return_val_if_fail (waiter != NULL, NULL);

	g_mutex_lock   (waiter->mutex);
	data = waiter->user_data;
	g_mutex_unlock (waiter->mutex);

	return data;
}
