/*-2 2002/08/19 18:05:43
 * $Id: rr-tcplistener.c,v 1.21 2003/02/25 15:57:18 jonas Exp $
 *
 * See the file LICENSE for redistribution information. 
 * If you have not received a copy of the license, please contact CodeFactory
 * by email at info@codefactory.se, or on the web at http://www.codefactory.se/
 * You may also write to: CodeFactory AB, SE-903 47, Ume, Sweden.
 *
 * Copyright (c) 2002 Jonas Borgstrm <jonas@codefactory.se>
 * Copyright (c) 2002 Daniel Lundin   <daniel@codefactory.se>
 * Copyright (c) 2002 CodeFactory AB.  All rights reserved.
 */

#include <librr/rr.h>

#ifdef G_OS_WIN32
#include "rr-win32.h"
#else
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <config.h>

#define BACKLOG 5

struct _RRTCPListenerPriv {
	GIOChannel *iochannel;
	
	GStaticMutex in_active;
	guint in_event;
	
	GStaticMutex err_active;
	guint err_event;
};

static GObjectClass *parent_class = NULL;

static void finalize (GObject *object);
static gboolean tcp_shutdown (RRListener *listener, GError **error);

static void
rr_tcp_listener_init (GObject *object)
{
	RRTCPListener *tcpl = (RRTCPListener *)object;
	RRTCPListenerPriv *priv;
	
	priv = g_new0 (RRTCPListenerPriv, 1);
	tcpl->priv = priv;

	g_static_mutex_init (&priv->in_active);
	g_static_mutex_init (&priv->err_active);
}

static void
rr_tcp_listener_class_init (GObjectClass *klass)
{
	RRListenerClass *listener_class = (RRListenerClass *)klass;
	klass->finalize = finalize;

	listener_class->shutdown = tcp_shutdown;

	parent_class = g_type_class_peek_parent (klass);
}

GType 
rr_tcp_listener_get_type (void)
{
	static GType rr_type = 0;

	if (!rr_type) {
		static GTypeInfo type_info = {
			sizeof (RRTCPListenerClass),
			NULL,
			NULL,
			(GClassInitFunc) rr_tcp_listener_class_init,
			NULL,
			NULL,
			sizeof (RRTCPListener),
			16,
			(GInstanceInitFunc) rr_tcp_listener_init
		};
		rr_type = g_type_register_static (RR_TYPE_LISTENER, 
						  "RRTCPListener", 
						  &type_info, 0);
#ifdef G_OS_WIN32
	rr_win32_init_winsock ();
#endif
	}
	return rr_type;
}

static gboolean
in_event (GIOChannel *source, GIOCondition condition, gpointer data)
{
	RRListener *listener = RR_LISTENER (data);
	RRTCPConnection *tcpc;
        struct sockaddr_in inet_addr;
	gint max_connections, num_connections;
#ifdef G_OS_WIN32
	int addrlen;
#else
        socklen_t addrlen;
#endif
        gint fd, sock;

	g_return_val_if_fail (condition == G_IO_IN, FALSE);

	sock = g_io_channel_unix_get_fd (source);
        
        addrlen = sizeof(struct sockaddr_in);
        /* Accept the connection */
        if ((fd = accept(sock, (struct sockaddr *)&inet_addr, 
			 &addrlen)) < 0) {

		g_warning ("accept () failed\n");
                return TRUE;
        }

	max_connections = rr_listener_get_max_connections (listener);
	num_connections = rr_listener_get_num_connections (listener);
	/* FIXME: we should probably send an error message here */
	if (max_connections != -1 &&
	    num_connections >= max_connections) {

		close (fd);
		return TRUE;
	}
	tcpc = rr_tcp_connection_new_unconnected (NULL);

	rr_listener_add_connection (listener, RR_CONNECTION (tcpc));

	rr_tcp_connection_connect_fd (tcpc, fd, RR_ROLE_LISTENER, NULL);
	g_object_unref (G_OBJECT (tcpc));

	return TRUE;
}

static gboolean
error_event (GIOChannel *source, GIOCondition condition, gpointer data)
{
	rr_debug2 ("listener::error_event: condition = %0x\n", condition);

	return FALSE;
}

static void
in_removed (gpointer data)
{
	RRTCPListener *tcpl = RR_TCP_LISTENER (data);

	g_static_mutex_unlock (&tcpl->priv->in_active);
}

static void
err_removed (gpointer data)
{
	RRTCPListener *tcpl = RR_TCP_LISTENER (data);

	g_static_mutex_unlock (&tcpl->priv->err_active);
}

/**
 * rr_tcp_listener_new:
 * @profreg: a list of profiles this listener should support.
 * @hostname: hostname to bind to
 * @port: TCP/IP port to listen on.
 * @error: location to return an error of type G_IO_ERROR, RR_ERROR or 
 * RR_BEEP_ERROR.
 * 
 * Creates a new RRTCPListener instance that listens for new connection
 * on address @hostname and port @port.
 *
 * NOTE: #RRTCPListener will decrease the refcount on @profreg when the
 * listener object is finalized.
 * 
 * Return value: a new #RRTCPListener object
 **/
RRListener *
rr_tcp_listener_new (RRProfileRegistry *profreg,
		     const gchar *hostname, 
		     gint port, GError **error)
{
	RRListener *listener;
	RRTCPListener *tcpl;
	RRTCPListenerPriv *priv;
#ifdef WITH_IPV6
	struct addrinfo *res, *ai, hints;
#else
	struct hostent *he;
        struct in_addr *haddr;
        struct sockaddr_in saddr;
	gint idx;
#endif
	gint fd, ret;
	gint one = 1;

#ifdef WITH_IPV6
	memset (&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_flags = AI_PASSIVE;
	ret = getaddrinfo (hostname, NULL, &hints, &res);
	if (ret != 0) {
		g_set_error (error,
			     RR_ERROR,                /* error domain */
			     RR_ERROR_GETADDRINFO,    /* error code */
			     gai_strerror (ret));     /* error message */
		return FALSE;
	}
	ai = res;
	while (ai) {
		fd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
		if (fd == -1 && ai->ai_next == NULL) {
			g_set_error (error,
				     RR_ERROR,                /* error domain */
				     RR_ERROR_SOCKET,         /* error code */
				     "socket() failed");      /* error message */
			freeaddrinfo (res);
			return NULL;
		}
		else if (fd < 0) {
			/* 
			 * Ok, socket() faild, but we'll try the rest of
			 * the matches
			 */
			ai = ai->ai_next;
			continue;
		}
		((struct sockaddr_in *)ai->ai_addr)->sin_port = htons(port);
		setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof (one));
		
		ret = bind(fd, ai->ai_addr, ai->ai_addrlen);
			
		if (ret < 0 && ai->ai_next == NULL) {
			g_set_error (error,
				     RR_ERROR,         /* error domain */
				     RR_ERROR_BIND,    /* error code */
				     "bind() failed"); /* error message */
			close (fd);
			freeaddrinfo (res);
			return FALSE;
		}
		if (ret == 0) {
			break;
		}
		close (fd);
		ai = ai->ai_next;
	}
	freeaddrinfo (res);
#else
        he = gethostbyname(hostname);
        if (he == NULL) {
		g_set_error (error,
			     RR_ERROR,                /* error domain */
			     RR_ERROR_GETHOSTBYNAME,  /* error code */
			     "gethostbyname() failed"); /* error message */
		return NULL;
        }
	idx = 0;
	for (;;) {
		haddr = ((struct in_addr *) (he->h_addr_list)[idx]);

		fd = socket(AF_INET, SOCK_STREAM, 0);
		if (fd == -1 && he->h_addr_list[idx + 1] == NULL) {
			g_set_error (error,
				     RR_ERROR,                /* error domain */
				     RR_ERROR_SOCKET,         /* error code */
				     "socket() failed");      /* error message */
			return NULL;
		}
		else if (fd == -1) {
			/* 
			 * Ok, socket() faild, but we'll try the rest of
			 * the matches
			 */
			idx++;
			continue;
		}
		setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof (one));

		memset(&saddr, 0, sizeof(struct sockaddr_in));
		saddr.sin_family = AF_INET;
		saddr.sin_port = htons(port);
		memcpy(&saddr.sin_addr, haddr, sizeof(struct in_addr));

		ret = bind(fd, (struct sockaddr *)&saddr, 
			   sizeof (struct sockaddr_in));
		if (ret < 0 && he->h_addr_list[idx + 1] == NULL) {
		
			close (fd);
			g_set_error (error,
				     RR_ERROR,          /* error domain */
				     RR_ERROR_BIND,     /* error code */
				     "bind() failed");  /* error message */
			
			return NULL;
		}
		if (ret == 0) {
			break;
		}
		idx++;
        }
#endif 
	if (listen(fd, BACKLOG) < 0) {

		g_set_error (error,
			     RR_ERROR,                /* error domain */
			     RR_ERROR_LISTEN,         /* error code */
			     "listen() failed");      /* error message */
		
                return NULL;
        }

	tcpl = g_object_new (RR_TYPE_TCP_LISTENER, NULL);
	listener = RR_LISTENER (tcpl);
	priv = tcpl->priv;

	if (profreg) {
		rr_listener_set_profile_registry (listener, profreg);
		g_object_unref (G_OBJECT (profreg));
	}

	priv->iochannel = g_io_channel_unix_new (fd);
	g_io_channel_set_close_on_unref (priv->iochannel, TRUE);
	g_io_channel_set_encoding (priv->iochannel, NULL, NULL);

	g_static_mutex_lock (&priv->err_active);
	priv->err_event = rr_add_watch_full (priv->iochannel, 0,
					     G_IO_HUP | G_IO_ERR,
					     error_event, tcpl,
					     err_removed);

	g_static_mutex_lock (&priv->in_active);
	priv->in_event = rr_add_watch_full (priv->iochannel, 0,
					    G_IO_IN, 
					    in_event, tcpl,
					    in_removed);

	rr_debug2 ("listener::listen ('%s', %d)", hostname, port);

	return listener;
}

/**
 * rr_tcp_listener_unlisten:
 * @tcpl: A #RRTCPListener object
 * @error: location to return an error of type G_IO_ERROR, RR_ERROR or 
 * RR_BEEP_ERROR.
 * 
 * Stop listening for new connections.
 *
 * NOTE: the #RRTCPListener object isn't unrefed.
 * 
 * Return value: %TRUE on success, %FALSE on failure.
 **/
gboolean
rr_tcp_listener_unlisten (RRTCPListener *tcpl, GError **error)
{
	RRTCPListenerPriv *priv;

	g_return_val_if_fail (RR_IS_TCP_LISTENER (tcpl), FALSE);

	priv = tcpl->priv;

	if (priv->in_event) {
		rr_source_remove (priv->in_event);
		priv->in_event = 0;
	}
	if (priv->err_event) {
		rr_source_remove (priv->err_event);
		priv->err_event = 0;
	}
	if (priv->iochannel) {
		g_io_channel_unref (priv->iochannel);
		priv->iochannel = NULL;
	}
	return TRUE;
}

static gboolean
tcp_shutdown (RRListener *listener, GError **error)
{
	RRTCPListener *tcpl = RR_TCP_LISTENER (listener);

	if (!rr_tcp_listener_unlisten (tcpl, error))
		return FALSE;

	return RR_LISTENER_CLASS (parent_class)->shutdown (listener, error);
}

static void
finalize (GObject *object)
{
	RRTCPListener *tcpl = RR_TCP_LISTENER (object);
	RRTCPListenerPriv *priv = tcpl->priv;

	rr_tcp_listener_unlisten (tcpl, NULL);

	g_static_mutex_free (&priv->in_active);
	g_static_mutex_free (&priv->err_active);
	g_free (priv);

	parent_class->finalize (object);
}
